from typing import ClassVar, Optional

from neontology.utils import get_rels_by_source, get_node_types

from neontology.basenode import BaseNode
from neontology.baserelationship import BaseRelationship


def test_get_rels_by_source():
    class AbstractNodeType(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        pp: str

    class MyNodeType(AbstractNodeType):
        __primarylabel__: ClassVar[Optional[str]] = "MyNodeType"

    class MyRelType(BaseRelationship):
        source: AbstractNodeType
        target: MyNodeType
        __relationshiptype__: ClassVar[Optional[str]] = "MY_REL_TYPE"

    rels_by_source = get_rels_by_source()

    assert rels_by_source["MyNodeType"] == {"MY_REL_TYPE"}


def test_no_primary_label_get_node_types():
    class SpecialPracticeNode(BaseNode):
        __primaryproperty__: ClassVar[str] = "pp"
        pp: str

    node_types = get_node_types()

    assert isinstance(node_types, dict)

    assert "SpecialPracticeNode" not in node_types.keys()



from abc import ABC, abstractmethod
from datetime import date, datetime, time, timedelta
from typing import Any, ClassVar, Dict, List, Optional, Set

from neo4j.time import Date as Neo4jDate
from neo4j.time import DateTime as Neo4jDateTime
from neo4j.time import Time as Neo4jTime
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    field_validator,
    model_validator,
)


class CommonModel(BaseModel, ABC):
    model_config = ConfigDict(
        validate_assignment=True,
        extra="forbid",
        arbitrary_types_allowed=True,
    )

    created: datetime = Field(
        default_factory=datetime.now, json_schema_extra={"set_on_create": True}
    )
    merged: Optional[datetime] = Field(default=None, validate_default=True)

    _set_on_match: List[str] = PrivateAttr()
    _set_on_create: List[str] = PrivateAttr()
    _always_set: List[str] = PrivateAttr()

    _neo4j_supported_types: ClassVar[Any] = (
        list,
        bool,
        int,
        bytearray,
        float,
        str,
        bytes,
        date,
        time,
        datetime,
        timedelta,
    )

    def __init__(self, **data: dict):
        super().__init__(**data)

        self._set_on_match = self._get_prop_usage("set_on_match")
        self._set_on_create = self._get_prop_usage("set_on_create")
        self._always_set = [
            x
            for x in self.model_dump().keys()
            if x not in self._set_on_match + self._set_on_create + ["source", "target"]
        ]

    @classmethod
    def _get_prop_usage(cls, usage_type: str) -> List[str]:
        all_props = cls.model_json_schema()["properties"]

        selected_props = []

        for prop, entry in all_props.items():
            if entry.get(usage_type) is True:
                selected_props.append(prop)

        return selected_props

    def _get_prop_values(
        self, props: List[str], exclude: Set[str] = set()
    ) -> Dict[str, Any]:
        """

        Returns:
            Dict[str, Any]: a dictionary of key/value pairs.
        """

        prop_values = {
            k: v for k, v in self.neo4j_dict(exclude=exclude).items() if k in props
        }

        return prop_values

    @abstractmethod
    def _get_merge_parameters(self) -> Dict[str, Any]:
        raise NotImplementedError

    @classmethod
    def export_type_converter(cls, value: Any) -> Any:
        if isinstance(value, dict):
            raise TypeError("Neo4j doesn't support dict types for properties.")

        elif isinstance(value, (tuple, set)):
            new_value = list(value)
            return cls.export_type_converter(new_value)

        elif isinstance(value, list):
            # items in a list must all be the same type
            item_type = type(value[0])
            for item in value:
                if isinstance(item, item_type) is False:
                    raise TypeError(
                        "For neo4j, all items in a list must be of the same type."
                    )

            return [cls.export_type_converter(x) for x in value]

        elif isinstance(value, cls._neo4j_supported_types) is False:
            return str(value)

        else:
            return value

    @classmethod
    def _export_dict_converter(cls, original_dict: Dict[str, Any]) -> Dict[str, Any]:
        """_summary_

        Args:
            export_dict (Dict[str, Any]): _description_

        Returns:
            Dict[str, Any]: _description_
        """

        export_dict = original_dict.copy()

        for k, v in export_dict.items():
            export_dict[k] = cls.export_type_converter(v)

        return export_dict

    def neo4j_dict(self, **kwargs: Dict[str, Any]) -> Dict[str, Any]:
        """Return a dict made up of only types compatible with neo4j

        Returns:
            dict: a dictionary export of this model instance
        """

        export_dict = self.model_dump(exclude_none=True, **kwargs)

        export_dict = self._export_dict_converter(export_dict)

        return export_dict

    #
    # validators
    #

    @field_validator("merged")
    def set_merged_to_created(
        cls, value: Optional[datetime], values: Dict[str, Any]
    ) -> datetime:
        """By default, set the 'merged' time equal to the 'created' time.

        If the 'merged' value has been explicitly set, this is preserved.

        Args:
            value (Optional[datetime]): the value of the field.
            values (Dict[str, Any]): a dictionary of field/value pairs set so far.

        Returns:
            datetime: The merged datetime value.
        """

        if value is None:
            return values.data["created"]
        else:
            return value

    @model_validator(mode="before")
    @classmethod
    def neo4j_datetime_to_native(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Datetimes come back from Neo4j as a non standard DateTime type.

        We check for any values where that is the case and convert them to
            native Python datetimes.

        See https://neo4j.com/docs/api/python-driver/4.4/temporal_types.html for further info.

        Args:
            values (Dict[str, Any]): Dictionary of field/value pairs from pydantic.

        Returns:
            Dict[str, Any]: Returns the dictionary, with any Neo4jDateTimes updated.
        """

        if not isinstance(values, dict):
            raise ValueError

        for key in values:
            if isinstance(values[key], (Neo4jDateTime, Neo4jDate, Neo4jTime)):
                values[key] = values[key].to_native()

        return values



"""Defines the BaseRelationship class.

The BaseRelationship class is used for creating and matching on relationships in the graph.

    Typical usage example:

    class MyRel(BaseRelationship):

        __relationshiptype__: ClassVar[Optional[str]] = "MY_REL"

        source: SourceNode
        target: TargetNode

    my_rel = MyRel(source=source_node, target=target_node)
    my_rel.merge()

"""

from typing import Any, ClassVar, Dict, List, Optional, Type, TypeVar

import numpy as np
import pandas as pd
from pydantic import PrivateAttr

from neontology.graphconnection import GraphConnection

from .basenode import BaseNode
from .commonmodel import CommonModel

R = TypeVar("R", bound="BaseRelationship")


class BaseRelationship(CommonModel):  # pyre-ignore[13]
    source: BaseNode
    target: BaseNode

    __relationshiptype__: ClassVar[Optional[str]] = None

    _merge_on: List[
        str
    ] = PrivateAttr()  # what relationship properties should we merge on

    def __init__(self, **data: dict):
        super().__init__(**data)

        self._merge_on = self._get_prop_usage("merge_on")

        # we can define 'abstract' relationships which don't have a label
        # these are to provide common properties to be used by subclassed relationships
        # but shouldn't be put in the graph or even instantiated
        if self.__relationshiptype__ is None:
            raise NotImplementedError(
                "Nodes to be used in the graph must define a primary label."
            )

    @classmethod
    def get_relationship_type(cls) -> str:
        """Get the relationship type to use for creating and matching this relationship.

        If __relationship__ has been specified, use that.

        Otherwise use the class name in uppercase

        Returns:
            str: the string to use for creating and matching this relationship
        """
        return cls.__relationshiptype__  # pyre-ignore[7]

    def _get_merge_parameters(
        self, source_prop: str, target_prop: str
    ) -> Dict[str, Any]:
        """

        Returns:
            Dict[str, Any]: a dictionary of key/value pairs.
        """

        exclusions = {"source", "target"}

        # these properties will be referenced individually
        merge_props = self._get_prop_values(self._merge_on, exclude=exclusions)

        params = {
            "source_prop": self.source.neo4j_dict()[source_prop],
            "target_prop": self.target.neo4j_dict()[target_prop],
            "always_set": self._get_prop_values(self._always_set, exclude=exclusions),
            "set_on_match": self._get_prop_values(
                self._set_on_match, exclude=exclusions
            ),
            "set_on_create": self._get_prop_values(
                self._set_on_create, exclude=exclusions
            ),
            **merge_props,
        }

        return params

    def merge(
        self,
    ) -> None:
        """Merge this relationship into the database."""
        source_label = self.source.__primarylabel__
        target_label = self.target.__primarylabel__

        source_pp = self.source.__primaryproperty__
        target_pp = self.target.__primaryproperty__

        params = self._get_merge_parameters(
            source_prop=source_pp, target_prop=target_pp
        )

        rel_type = self.get_relationship_type()

        # build a string of properties to merge on "prop_name: $prop_name"
        merge_props = ", ".join([f"{x}: ${x}" for x in self._merge_on])

        cypher = f"""
        MATCH (source:{source_label} {{ {source_pp}: $source_prop }}),
            (target:{target_label} {{ {target_pp}: $target_prop }})
        MERGE (source)-[r:{rel_type} {{ {merge_props} }}]->(target)
        ON MATCH SET r += $set_on_match
        ON CREATE SET r += $set_on_create
        SET r += $always_set
        """

        graph = GraphConnection()

        graph.cypher_write(cypher, params)

    @classmethod
    def merge_relationships(
        cls: Type[R],
        rels: List[R],
        source_type: Optional[Type[BaseNode]] = None,
        target_type: Optional[Type[BaseNode]] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Merge multiple relationships (of this type) into the database.

        Sometimes the source and target label may be ambiguous (e.g. where we have subclassed nodes)
            In this case you can explicitly pass in the relevant types

        Sometimes we want to match nodes on a property which isn't the primary property, so we can specify what property to use.

        Args:
            cls (Type[R]): this class
            rels (List[R]): a list of relationships which are instances of this class

        Raises:
            TypeError: If relationships are provided which aren't of this class
        """

        if source_type is None:
            source_type = cls.model_fields["source"].annotation

        if target_type is None:
            target_type = cls.model_fields["target"].annotation

        for rel in rels:
            if isinstance(rel, cls) is False:
                raise TypeError("Relationship was incorrect type.")
            if type(rel.source) is not source_type:
                raise TypeError("Received an inappropriate kind of source node.")
            if type(rel.target) is not target_type:
                raise TypeError("Received an inappropriate kind of target node.")

        if source_prop is None:
            source_prop = source_type.__primaryproperty__

        if target_prop is None:
            target_prop = target_type.__primaryproperty__

        source_label = source_type.__primarylabel__
        target_label = target_type.__primarylabel__

        # build a string of properties to merge on "prop_name: $prop_name"
        # we need to instantiate the class so that _merge_on is generated as part of __init__
        merge_props = ", ".join([f"{x}: ${x}" for x in cls._get_prop_usage("merge_on")])

        rel_list: List[Dict[str, Any]] = [
            x._get_merge_parameters(source_prop, target_prop) for x in rels
        ]

        rel_type = cls.get_relationship_type()

        cypher = f"""
        UNWIND $rel_list AS rel
        MATCH (source:{source_label})
        WHERE source.{source_prop} = rel.source_prop
        MATCH (target:{target_label})
        WHERE target.{target_prop} = rel.target_prop
        MERGE (source)-[r:{rel_type} {{ {merge_props} }}]->(target)
        ON MATCH SET r += rel.set_on_match
        ON CREATE SET r += rel.set_on_create
        SET r += rel.always_set
        """

        graph = GraphConnection()

        graph.cypher_write(cypher=cypher, params={"rel_list": rel_list})

    @classmethod
    def merge_records(
        cls: Type[R],
        records: List[Dict[str, Any]],
        source_type: Optional[Type[BaseNode]] = None,
        target_type: Optional[Type[BaseNode]] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Take a list of dictionaries and use them to merge in relationships in the graph.

        Sometimes, a relationship can accept nodes which subclass a particular node type.
            In these instances, it may be necessary to explicitly state what type of node should be used.

        Each record should have a source and target key where the value is the primary property
            value of the respective nodes.

        Args:
            records (List[Dict[str, Any]]): a list of dictionaries used to populate relationships
            source_type: explicitly state the class to use for source node
            target_type: explicitly state the class to use for target node
        """

        hydrated_list = []

        if source_type is None:
            source_type = cls.model_fields["source"].annotation

        if target_type is None:
            target_type = cls.model_fields["target"].annotation

        if source_prop is None:
            source_prop = source_type.__primaryproperty__

        if target_prop is None:
            target_prop = target_type.__primaryproperty__

        for record in records:
            hydrated = dict(record)

            hydrated["source"] = source_type.model_construct(
                **{source_prop: record["source"]}
            )
            hydrated["target"] = target_type.model_construct(
                **{target_prop: record["target"]}
            )

            hydrated_list.append(hydrated)

        rels = [cls(**x) for x in hydrated_list]

        cls.merge_relationships(
            rels,
            source_type=source_type,
            source_prop=source_prop,
            target_type=target_type,
            target_prop=target_prop,
        )

    @classmethod
    def merge_df(
        cls: Type[R],
        df: pd.DataFrame,
        source_type: Optional[Type[BaseNode]] = None,
        target_type: Optional[Type[BaseNode]] = None,
        source_prop: Optional[str] = None,
        target_prop: Optional[str] = None,
    ) -> None:
        """Merge in relationships based on data in a pandas data frame

        Expects columns named 'source' and 'target' with the primary property value
            for the source and target nodes.

        Then additional fields should have a corresponding column.

        Args:
            df (pd.DataFrame): pandas dataframe where each row represents a relationship to merge
        """

        if df.empty is False:
            records = df.replace([np.nan], None).to_dict(orient="records")
            cls.merge_records(
                records,
                source_type=source_type,
                source_prop=source_prop,
                target_type=target_type,
                target_prop=target_prop,
            )



import os
from typing import Any, Dict, List, Optional

from dotenv import load_dotenv
from neo4j import GraphDatabase, Neo4jDriver
from neo4j import Record as Neo4jRecord
from neo4j import Transaction as Neo4jTransaction


class GraphConnection(object):
    """Class for managing connections to Neo4j."""

    _instance = None

    def __new__(
        cls,
        neo4j_uri: Optional[str] = None,
        neo4j_username: Optional[str] = None,
        neo4j_password: Optional[str] = None,
    ) -> "GraphConnection":
        """Make sure we only have a single connection to the GraphDatabase.

        This connection then gets used by all instances.

        Args:
            neo4j_uri (Optional[str], optional): Neo4j URI to connect to. Defaults to None.
            neo4j_username (Optional[str], optional): Neo4j username. Defaults to None.
            neo4j_password (Optional[str], optional): Neo4j password. Defaults to None.

        Returns:
            GraphConnection: Instance of the connection
        """

        if cls._instance is None:
            cls._instance = object.__new__(cls)

            if GraphConnection._instance:
                try:
                    driver = GraphConnection._instance.driver = GraphDatabase.driver(  # type: ignore
                        neo4j_uri, auth=(neo4j_username, neo4j_password)
                    )
                    driver.verify_connectivity()

                except Exception as error:
                    print(
                        "Error: connection not established. Have you run init_neontology? {}".format(
                            error
                        )
                    )
                    GraphConnection._instance = None

            else:
                GraphConnection._instance = None

        return cls._instance

    def __del__(self) -> None:
        """Close the driver gracefully when the class gets deleted."""

        self.driver.close()

    def __init__(
        self,
        neo4j_uri: Optional[str] = None,
        neo4j_username: Optional[str] = None,
        neo4j_password: Optional[str] = None,
    ) -> None:
        if self._instance:
            self.driver: Neo4jDriver = self._instance.driver

    def run_transaction_single(
        self, tx: Neo4jTransaction, query: str, params: Dict[str, Any]
    ) -> Optional[Neo4jRecord]:
        """Run a transaction which is expected to return a single result.

        Args:
            tx (Neo4jTransaction): Neo4j Transaction object
            query (str): cypher query to run
            params (Dict[str, Any]): Parameters to pass to the query

        Returns:
            Optional[Neo4jRecord]: The result
        """

        return tx.run(query, **params).single()

    def run_transaction_many(
        self, tx: Neo4jTransaction, query: str, params: Dict[str, Any]
    ) -> List[Neo4jRecord]:
        """Run a transation which is expected to return multiple nodes.

        Args:
            tx (Neo4jTransaction): Neo4j Transaction object
            query (str): cypher query to run
            params (Dict[str, Any]): parameters to pass the query

        Returns:
            List[Neo4jRecord]: a list of the results
        """

        return [record for record in tx.run(query, **params)]

    def cypher_write(self, cypher: str, params: Dict[str, Any] = {}) -> None:
        """Execute a write transaction.

        Args:
            cypher (str): cypher query
            params (Dict[str, Any]): parameters to pass to the query
        """

        with self.driver.session() as session:
            session.execute_write(self.run_transaction_single, cypher, params)

    def cypher_write_single(self, cypher: str, params: Dict[str, Any] = {}) -> None:
        """Execute a write transaction.

        Args:
            cypher (str): cypher query
            params (Dict[str, Any]): parameters to pass to the query
        """

        with self.driver.session() as session:
            return session.execute_write(self.run_transaction_single, cypher, params)

    def cypher_write_many(self, cypher: str, params: Dict[str, Any] = {}) -> None:
        """Execute a write transaction.

        Args:
            cypher (str): cypher query
            params (Dict[str, Any]): parameters to pass to the query
        """

        with self.driver.session() as session:
            return session.execute_write(self.run_transaction_many, cypher, params)

    def cypher_read(
        self, cypher: str, params: Dict[str, Any] = {}
    ) -> Optional[Neo4jRecord]:
        """Run a cypher read only query which is expected to return a single result.

        Args:
            cypher (str): cypher query string
            params (Dict[str, Any]): parameters to pass to the query

        Returns:
            Neo4jRecord: the resulting Neo4j 'Record', or None
        """

        with self.driver.session() as session:
            return session.execute_read(self.run_transaction_single, cypher, params)

    def cypher_read_many(
        self, cypher: str, params: Dict[str, Any] = {}
    ) -> List[Neo4jRecord]:
        """Run a cypher read query which will return multiple records.

        Args:
            cypher (str): cypher string to run
            params (Dict[str, Any]): parameters to pass to the query

        Returns:
            List[Neo4jRecord]: A list of Neo4j 'Records' returned by the query.
        """

        with self.driver.session() as session:
            return session.execute_read(self.run_transaction_many, cypher, params)

    def apply_constraint(self, label: str, property: str) -> None:
        cypher = f"""
        CREATE CONSTRAINT IF NOT EXISTS
        FOR (n:{label})
        REQUIRE n.{property} IS UNIQUE
        """

        self.cypher_write(cypher)


def init_neontology(
    neo4j_uri: Optional[str] = None,
    neo4j_username: Optional[str] = None,
    neo4j_password: Optional[str] = None,
) -> None:
    """Initialise neontology.

    If connection properties are explicitly passed in, use these.
    If not, attempt to load from enviornment variables (optionally in a .env file.)

    Args:
        neo4j_uri (Optional[str], optional): Neo4j URI to connect to. Defaults to None.
        neo4j_username (Optional[str], optional): Neo4j username. Defaults to None.
        neo4j_password (Optional[str], optional): Neo4j password. Defaults to None.
    """

    # try to load environment variables from .env file
    load_dotenv()

    if neo4j_uri is None:
        neo4j_uri = os.getenv("NEO4J_URI")

    if neo4j_password is None:
        neo4j_password = os.getenv("NEO4J_PASSWORD")

    if neo4j_username is None:
        neo4j_username = os.getenv("NEO4J_USERNAME")

    GraphConnection(neo4j_uri, neo4j_username, neo4j_password)
